@@ -76,6 +76,8 @@ enum {
 
 enum {
     SCREEN_CHANGED,
+    SCREEN_OUTPUT_CONNECTED,
+    SCREEN_OUTPUT_DISCONNECTED,
     SCREEN_SIGNAL_LAST,
 };
 
@@ -96,8 +98,10 @@ struct GnomeRROutput
     GnomeRRMode **	modes;
     int			n_preferred;
     guint8 *		edid_data;
-    int         edid_size;
+    gsize		edid_size;
     char *              connector_type;
+    gint		backlight_min;
+    gint		backlight_max;
 };
 
 struct GnomeRROutputWrap
@@ -560,6 +564,80 @@ screen_info_new (GnomeRRScreen *screen, gboolean needs_reprobe, GError **error)
     }
 }
 
+static GnomeRROutput *
+find_output_by_id (GnomeRROutput **haystack, guint32 id)
+{
+    guint i;
+
+    for (i = 0; haystack[i] != NULL; i++)
+    {
+	if (gnome_rr_output_get_id (haystack[i]) == id)
+	    return haystack[i];
+    }
+    return NULL;
+}
+
+static void
+diff_outputs_and_emit_signals (ScreenInfo *old, ScreenInfo *new)
+{
+    guint i;
+    guint32 id_old, id_new;
+    GnomeRROutput *output_old;
+    GnomeRROutput *output_new;
+
+    /* have any outputs been removed or disconnected */
+    for (i = 0; old->outputs[i] != NULL; i++)
+    {
+        id_old = gnome_rr_output_get_id (old->outputs[i]);
+        output_new = find_output_by_id (new->outputs, id_old);
+	if (output_new == NULL)
+	{
+	    /* output removed (and disconnected) */
+	    if (gnome_rr_output_is_connected (old->outputs[i]))
+	     {
+	        g_signal_emit (G_OBJECT (new->screen),
+			       screen_signals[SCREEN_OUTPUT_DISCONNECTED], 0,
+			       old->outputs[i]);
+             }
+	    continue;
+	}
+	if (gnome_rr_output_is_connected (old->outputs[i]) &&
+	    !gnome_rr_output_is_connected (output_new))
+	{
+	    /* output disconnected */
+	    g_signal_emit (G_OBJECT (new->screen),
+			   screen_signals[SCREEN_OUTPUT_DISCONNECTED], 0,
+			   old->outputs[i]);
+        }
+    }
+
+    /* have any outputs been created or connected */
+    for (i = 0; new->outputs[i] != NULL; i++)
+    {
+        id_new = gnome_rr_output_get_id (new->outputs[i]);
+        output_old = find_output_by_id (old->outputs, id_new);
+	if (output_old == NULL)
+	{
+	    /* output created */
+	    if (gnome_rr_output_is_connected (new->outputs[i]))
+	     {
+	        g_signal_emit (G_OBJECT (new->screen),
+			       screen_signals[SCREEN_OUTPUT_CONNECTED], 0,
+			       new->outputs[i]);
+            }
+	    continue;
+	}
+	if (!gnome_rr_output_is_connected (output_old) &&
+	    gnome_rr_output_is_connected (new->outputs[i]))
+	{
+	    /* output connected */
+	    g_signal_emit (G_OBJECT (new->screen),
+			   screen_signals[SCREEN_OUTPUT_CONNECTED], 0,
+			   new->outputs[i]);
+         }
+    }
+}
+
 static gboolean
 screen_update (GnomeRRScreen *screen, gboolean force_callback, gboolean needs_reprobe, GError **error)
 {
@@ -577,6 +655,9 @@ screen_update (GnomeRRScreen *screen, gboolean force_callback, gboolean needs_re
 	    changed = TRUE;
 #endif
 
+    /* work out if any outputs have changed connected state */
+    diff_outputs_and_emit_signals (screen->priv->info, info);
+
     screen_info_free (screen->priv->info);
 	
     screen->priv->info = info;
@@ -748,7 +829,8 @@ gnome_rr_screen_finalize (GObject *gobject)
 
     gdk_window_remove_filter (screen->priv->gdk_root, screen_on_event, screen);
 
-    screen_info_free (screen->priv->info);
+    if (screen->priv->info)
+      screen_info_free (screen->priv->info);
 
     G_OBJECT_CLASS (gnome_rr_screen_parent_class)->finalize (gobject);
 }
@@ -823,6 +905,63 @@ gnome_rr_screen_class_init (GnomeRRScreenClass *klass)
             g_cclosure_marshal_VOID__VOID,
             G_TYPE_NONE,
 	    0);
+
+    /**
+     * GnomeRRScreen::output-connected:
+     * @screen: the #GnomeRRScreen that emitted the signal
+     * @output: the #GnomeRROutput that was connected
+     *
+     * This signal is emitted when a display device is connected to a
+     * port, or a port is hotplugged with an active output. The latter
+     * can happen if a laptop is docked, and the dock provides a new
+     * active output.
+     *
+     * The @output value is not a #GObject. The returned @output value can
+     * only assume to be valid during the emission of the signal (i.e. within
+     * your signal handler only), as it may change later when the @screen
+     * is modified due to an event from the X server, or due to another
+     * place in the application modifying the @screen and the @output.
+     * Therefore, deal with changes to the @output right in your signal
+     * handler, instead of keeping the @output reference for an async or
+     * idle function.
+     **/
+    screen_signals[SCREEN_OUTPUT_CONNECTED] = g_signal_new("output-connected",
+            G_TYPE_FROM_CLASS (gobject_class),
+            G_SIGNAL_RUN_FIRST | G_SIGNAL_NO_RECURSE | G_SIGNAL_NO_HOOKS,
+            G_STRUCT_OFFSET (GnomeRRScreenClass, output_connected),
+            NULL,
+            NULL,
+            g_cclosure_marshal_VOID__POINTER,
+            G_TYPE_NONE,
+            1, G_TYPE_POINTER);
+
+    /**
+     * GnomeRRScreen::output-disconnected:
+     * @screen: the #GnomeRRScreen that emitted the signal
+     * @output: the #GnomeRROutput that was disconnected
+     *
+     * This signal is emitted when a display device is disconnected from
+     * a port, or a port output is hot-unplugged. The latter can happen
+     * if a laptop is undocked, and the dock provided the output.
+     *
+     * The @output value is not a #GObject. The returned @output value can
+     * only assume to be valid during the emission of the signal (i.e. within
+     * your signal handler only), as it may change later when the @screen
+     * is modified due to an event from the X server, or due to another
+     * place in the application modifying the @screen and the @output.
+     * Therefore, deal with changes to the @output right in your signal
+     * handler, instead of keeping the @output reference for an async or
+     * idle function.
+     **/
+    screen_signals[SCREEN_OUTPUT_DISCONNECTED] = g_signal_new("output-disconnected",
+            G_TYPE_FROM_CLASS (gobject_class),
+            G_SIGNAL_RUN_FIRST | G_SIGNAL_NO_RECURSE | G_SIGNAL_NO_HOOKS,
+            G_STRUCT_OFFSET (GnomeRRScreenClass, output_disconnected),
+            NULL,
+            NULL,
+            g_cclosure_marshal_VOID__POINTER,
+            G_TYPE_NONE,
+            1, G_TYPE_POINTER);
 }
 
 void
@@ -947,6 +1086,7 @@ gnome_rr_screen_get_timestamps (GnomeRRScreen *screen,
 static gboolean
 force_timestamp_update (GnomeRRScreen *screen)
 {
+#ifdef HAVE_RANDR
     GnomeRRScreenPrivate *priv = screen->priv;
     GnomeRRCrtc *crtc;
     XRRCrtcInfo *current_info;
@@ -989,6 +1129,9 @@ force_timestamp_update (GnomeRRScreen *screen)
 	timestamp_updated = TRUE;
 out:
     return timestamp_updated;
+#else
+    return FALSE;
+#endif
 }
 
 /**
@@ -1154,7 +1297,7 @@ static guint8 *
 get_property (Display *dpy,
 	      RROutput output,
 	      Atom atom,
-	      int *len)
+	      gsize *len)
 {
 #ifdef HAVE_RANDR
     unsigned char *prop;
@@ -1189,7 +1332,7 @@ get_property (Display *dpy,
 }
 
 static guint8 *
-read_edid_data (GnomeRROutput *output, int *len)
+read_edid_data (GnomeRROutput *output, gsize *len)
 {
     Atom edid_atom;
     guint8 *result;
@@ -1258,6 +1401,42 @@ out:
 #endif
 }
 
+static void
+update_brightness_limits (GnomeRROutput *output)
+{
+#ifdef HAVE_RANDR
+    gint rc;
+    Atom atom;
+    XRRPropertyInfo *info;
+
+    gdk_error_trap_push ();
+    atom = XInternAtom (DISPLAY (output), "BACKLIGHT", FALSE);
+    info = XRRQueryOutputProperty (DISPLAY (output), output->id, atom);
+    rc = gdk_error_trap_pop ();
+    if (rc != 0)
+    {
+        g_warning ("could not get output property for %s, rc: %i",
+		   output->name, rc);
+        goto out;
+    }
+    if (info == NULL)
+    {
+        g_warning ("could not get output property for %s",
+		   output->name);
+        goto out;
+    }
+    if (!info->range || info->num_values != 2)
+    {
+        g_debug ("backlight %s was not range", output->name);
+        goto out;
+    }
+    output->backlight_min = info->values[0];
+    output->backlight_max = info->values[1];
+out:
+    XFree (info);
+#endif
+}
+
 #ifdef HAVE_RANDR
 static gboolean
 output_initialize (GnomeRROutput *output, XRRScreenResources *res, GError **error)
@@ -1329,7 +1508,11 @@ output_initialize (GnomeRROutput *output, XRRScreenResources *res, GError **erro
     
     /* Edid data */
     output->edid_data = read_edid_data (output, &output->edid_size);
-    
+
+    /* brightness data */
+    if (output->connected)
+        update_brightness_limits (output);
+
     XRRFreeOutputInfo (info);
 
     return TRUE;
@@ -1354,6 +1537,8 @@ output_copy (const GnomeRROutput *from)
     output->connected = from->connected;
     output->n_preferred = from->n_preferred;
     output->connector_type = g_strdup (from->connector_type);
+    output->backlight_min = -1;
+    output->backlight_max = -1;
 
     array = g_ptr_array_new ();
     for (p_crtc = from->possible_crtcs; *p_crtc != NULL; p_crtc++)
@@ -1403,13 +1588,161 @@ gnome_rr_output_get_id (GnomeRROutput *output)
 }
 
 const guint8 *
-gnome_rr_output_get_edid_data (GnomeRROutput *output)
+gnome_rr_output_get_edid_data (GnomeRROutput *output, gsize *size)
 {
     g_return_val_if_fail (output != NULL, NULL);
-    
+    if (size)
+        *size = output->edid_size;
     return output->edid_data;
 }
 
+/**
+ * gnome_rr_output_get_backlight_min:
+ *
+ * Returns: The mimimum backlight value, or -1 if not supported
+ */
+gint
+gnome_rr_output_get_backlight_min (GnomeRROutput *output)
+{
+    g_return_val_if_fail (output != NULL, -1);
+    return output->backlight_min;
+}
+
+/**
+ * gnome_rr_output_get_backlight_max:
+ *
+ * Returns: The maximum backlight value, or -1 if not supported
+ */
+gint
+gnome_rr_output_get_backlight_max (GnomeRROutput *output)
+{
+    g_return_val_if_fail (output != NULL, -1);
+    return output->backlight_max;
+}
+
+/**
+ * gnome_rr_output_get_backlight:
+ *
+ * Returns: The currently set backlight brightness
+ */
+gint
+gnome_rr_output_get_backlight (GnomeRROutput *output, GError **error)
+{
+    guint now = -1;
+#ifdef HAVE_RANDR
+    unsigned long nitems;
+    unsigned long bytes_after;
+    guint *prop;
+    Atom atom;
+    Atom actual_type;
+    int actual_format;
+    gint retval;
+
+    g_return_val_if_fail (output != NULL, -1);
+
+    gdk_error_trap_push ();
+    atom = XInternAtom (DISPLAY (output), "BACKLIGHT", FALSE);
+    retval = XRRGetOutputProperty (DISPLAY (output), output->id, atom,
+				   0, 4, False, False, None,
+				   &actual_type, &actual_format,
+				   &nitems, &bytes_after, ((unsigned char **)&prop));
+    gdk_flush ();
+    if (gdk_error_trap_pop ())
+    {
+        g_set_error_literal (error,
+			     GNOME_RR_ERROR,
+			     GNOME_RR_ERROR_UNKNOWN,
+			     "unhandled X error while getting the range of backlight values");
+        goto out;
+    }
+
+    if (retval != Success) {
+        g_set_error_literal (error,
+			     GNOME_RR_ERROR,
+			     GNOME_RR_ERROR_RANDR_ERROR,
+			     "could not get the range of backlight values");
+        goto out;
+    }
+    if (actual_type == XA_INTEGER &&
+        nitems == 1 &&
+        actual_format == 32)
+    {
+        memcpy (&now, prop, sizeof (guint));
+    }
+    else
+    {
+	g_set_error (error,
+		     GNOME_RR_ERROR,
+		     GNOME_RR_ERROR_RANDR_ERROR,
+		     "failed to get correct property type, got %lu,%i",
+		     nitems, actual_format);
+    }
+out:
+    XFree (prop);
+#else
+    g_set_error_literal (error,
+			 GNOME_RR_ERROR,
+			 GNOME_RR_ERROR_NO_RANDR_EXTENSION,
+			 "not compiled with RANDR support");
+#endif /* HAVE_RANDR */
+    return now;
+}
+
+/**
+ * gnome_rr_output_set_backlight:
+ * @value: the absolute value which is min >= this <= max
+ *
+ * Returns: %TRUE for success
+ */
+gboolean
+gnome_rr_output_set_backlight (GnomeRROutput *output, gint value, GError **error)
+{
+    gboolean ret = FALSE;
+#ifdef HAVE_RANDR
+    Atom atom;
+
+    g_return_val_if_fail (output != NULL, FALSE);
+
+    /* check this is sane */
+    if (value < output->backlight_min ||
+        value > output->backlight_max)
+    {
+	g_set_error (error,
+		     GNOME_RR_ERROR,
+		     GNOME_RR_ERROR_BOUNDS_ERROR,
+		     "out of brightness range: %i, has to be %i -> %i",
+		     value,
+		     output->backlight_max, output->backlight_min);
+	goto out;
+    }
+
+    /* don't abort on error */
+    gdk_error_trap_push ();
+    atom = XInternAtom (DISPLAY (output), "BACKLIGHT", FALSE);
+    XRRChangeOutputProperty (DISPLAY (output), output->id, atom,
+			     XA_INTEGER, 32, PropModeReplace,
+			     (unsigned char *) &value, 1);
+    if (gdk_error_trap_pop ())
+    {
+        g_set_error_literal (error,
+			     GNOME_RR_ERROR,
+			     GNOME_RR_ERROR_UNKNOWN,
+			     "unhandled X error while setting the backlight values");
+        goto out;
+    }
+
+    /* we assume this succeeded as there's no return value */
+    ret = TRUE;
+#else
+    g_set_error_literal (error,
+			 GNOME_RR_ERROR,
+			 GNOME_RR_ERROR_NO_RANDR_EXTENSION,
+			 "not compiled with RANDR support");
+#endif /* HAVE_RANDR */
+out:
+    return ret;
+}
+
 /**
  * gnome_rr_screen_get_output_by_id:
  *
